--[[
  My ComputerCraft Package Manager
  Supports multiple repositories, dependency resolution, index and package caching.
]]

local CKG = {}

local INSTALL_DIR = "/ckg"
local CACHE_DIR = fs.combine(INSTALL_DIR, "cache")
local REPO_CACHE_DIR = fs.combine(CACHE_DIR, "repos")
local CONFIG_FILE = fs.combine(INSTALL_DIR, "config.txt")
local DEFAULT_REPOS = {
  "http://193.23.220.51:3000/p2vman/cpkg/raw/branch/main/packages.json"
}

local function load_config()
  if not fs.exists(CONFIG_FILE) then
    return { repos = DEFAULT_REPOS }
  end
  local f = fs.open(CONFIG_FILE, "r")
  local cfg = textutils.unserialize(f.readAll())
  f.close()
  return cfg
end

local function save_config(cfg)
  local f = fs.open(CONFIG_FILE, "w")
  f.write(textutils.serialize(cfg))
  f.close()
end

local function fetch_package_lists(cfg)
  if not fs.exists(REPO_CACHE_DIR) then fs.makeDir(REPO_CACHE_DIR) end
  local all = {}

  for i, url in ipairs(cfg.repos) do
    local cached_file = fs.combine(REPO_CACHE_DIR, "repo_" .. i .. ".json")

    if fs.exists(cached_file) then
      local f = fs.open(cached_file, "r")
      local data = textutils.unserializeJSON(f.readAll())
      f.close()
      for name, info in pairs(data) do
        all[name] = info
      end
    else
      if http.checkURL(url) then
        print("Fetching index from " .. url)
        local tmp = ".pkg_tmp_repo.json"
        if fs.exists(tmp) then fs.delete(tmp) end
        shell.run("wget", url, tmp)
        local f = fs.open(tmp, "r")
        local data = textutils.unserializeJSON(f.readAll())
        f.close()
        fs.delete(tmp)

        local cache = fs.open(cached_file, "w")
        cache.write(textutils.serializeJSON(data))
        cache.close()

        for name, info in pairs(data) do
          all[name] = info
        end
      else
        print("Warning: cannot reach repo " .. url)
      end
    end
  end

  return all
end

local function install_files(base_url, files, install_path)
  for _, file in ipairs(files) do
    local url = base_url .. "/" .. file
    local path = fs.combine(install_path, file)
    local dir = fs.getDir(path)
    if not fs.exists(dir) then fs.makeDir(dir) end
    if fs.exists(path) then fs.delete(path) end
    shell.run("wget", url, path)
  end
end

local function install_single(name, list, installed)
  if installed[name] then return end
  local pkg = list[name]
  if not pkg then
    error("Package not found: " .. name)
  end

  if pkg.dependencies then
    for _, dep in ipairs(pkg.dependencies) do
      install_single(dep, list, installed)
    end
  end

  local install_path = fs.combine(INSTALL_DIR, name)
  if fs.exists(install_path) then
    print("Already installed: " .. name)
    installed[name] = true
    return
  end

  if not fs.exists(CACHE_DIR) then fs.makeDir(CACHE_DIR) end
  local cache_path = fs.combine(CACHE_DIR, name)

  if fs.exists(cache_path) then
    print("Installing " .. name .. " from cache...")
    fs.copy(cache_path, install_path)
  else
    print("Downloading " .. name .. "...")
    fs.makeDir(cache_path)
    install_files(pkg.base_url, pkg.files, cache_path)
    fs.copy(cache_path, install_path)
  end

  print("Installed " .. name)
  installed[name] = true
end

function CKG.install(name)
  local cfg = load_config()
  local list = fetch_package_lists(cfg)
  install_single(name, list, {})
end

function CKG.fetch(name)
  local cfg = load_config()
  local list = fetch_package_lists(cfg)
  local pkg = list[name]
  if not pkg then
    error("Package not found: " .. name)
  end
  if not fs.exists(CACHE_DIR) then fs.makeDir(CACHE_DIR) end
  local cache_path = fs.combine(CACHE_DIR, name)
  if fs.exists(cache_path) then fs.delete(cache_path) end
  fs.makeDir(cache_path)
  install_files(pkg.base_url, pkg.files, cache_path)
  print("Fetched " .. name .. " to cache")
end

function CKG.remove(name)
  local install_path = fs.combine(INSTALL_DIR, name)
  if fs.exists(install_path) then
    fs.delete(install_path)
    print("Removed " .. name)
  else
    print("Package not installed: " .. name)
  end
end

function CKG.list()
  if not fs.exists(INSTALL_DIR) then
    print("No packages installed.")
    return
  end
  local list = fs.list(INSTALL_DIR)
  for _, pkg in ipairs(list) do
    if pkg ~= "cache" and pkg ~= "config.txt" then
      print(pkg)
    end
  end
end

function CKG.add_repo(url)
  local cfg = load_config()
  table.insert(cfg.repos, url)
  save_config(cfg)
  print("Repository added: " .. url)
end

function CKG.list_repos()
  local cfg = load_config()
  for _, url in ipairs(cfg.repos) do
    print(url)
  end
end

function CKG.clear_repos()
  save_config({ repos = DEFAULT_REPOS })
  print("Repositories reset to default")
end

function CKG.clear_cache()
  if fs.exists(CACHE_DIR) then
    fs.delete(CACHE_DIR)
    print("Cache cleared")
  else
    print("Cache is already empty")
  end
end

function CKG.update()
  if fs.exists(REPO_CACHE_DIR) then fs.delete(REPO_CACHE_DIR) end
  fs.makeDir(REPO_CACHE_DIR)

  local cfg = load_config()
  for i, url in ipairs(cfg.repos) do
    if http.checkURL(url) then
      print("Updating index from " .. url)
      local tmp = ".pkg_tmp_repo.json"
      if fs.exists(tmp) then fs.delete(tmp) end
      shell.run("wget", url, tmp)
      local f = fs.open(tmp, "r")
      local data = textutils.unserializeJSON(f.readAll())
      f.close()
      fs.delete(tmp)

      local cached_file = fs.combine(REPO_CACHE_DIR, "repo_" .. i .. ".json")
      local cache = fs.open(cached_file, "w")
      cache.write(textutils.serializeJSON(data))
      cache.close()
    else
      print("Warning: cannot reach repo " .. url)
    end
  end

  print("Repository indexes updated!")
end

-- CLI interface
local args = { ... }
local cmd = args[1]

if cmd == "install" then
    CKG.install(args[2])
elseif cmd == "remove" then
    CKG.remove(args[2])
elseif cmd == "list" then
    CKG.list()
elseif cmd == "fetch" then
    CKG.fetch(args[2])
elseif cmd == "clear-cache" then
    CKG.clear_cache()
elseif cmd == "add-repo" then
    CKG.add_repo(args[2])
elseif cmd == "list-repos" then
    CKG.list_repos()
elseif cmd == "clear-repos" then
    CKG.clear_repos()
elseif cmd == "update" then
    CKG.update()
else
  print([[
Usage:
  ckg install <name>       Install a package
  ckg remove <name>        Remove a package
  ckg list                 List installed packages
  ckg fetch <name>         Download package to cache
  ckg clear-cache          Clear downloaded packages cache
  ckg add-repo <url>       Add a new repository
  ckg list-repos           List repositories
  ckg clear-repos          Reset repositories to default
  ckg update               Refresh package indexes
]])
end
